use std::{
    collections::HashMap,
    io::{Read, Write},
};

use varint_rs::{VarintReader, VarintWriter};

use crate::{
    Connected, Server, Status,
    error::{ChatError, Result},
    readers::{read_packet, read_plugin_msg, read_string, try_read_plugin_msg},
    status::ServerType,
    writers::{write_packet, write_plugin_msg, write_string},
};

impl Server<Status> {
    pub fn connect(self) -> Result<Server<Connected>> {
        let Self {
            mut stream,
            addr,
            port,
            lang,
            state:
                Status {
                    typ,
                    description,
                    mods,
                },
        } = self;

        let (uuid, username) = match typ {
            ServerType::Vanilla => {
                // Handshake
                write_packet(&mut stream, |buf| {
                    buf.write_u8_varint(0)?;
                    buf.write_u8_varint(5)?;
                    buf.write_usize_varint(addr.len())?;
                    buf.write_all(addr.as_bytes())?;
                    buf.write_all(&port.to_le_bytes())?;
                    buf.write_u8_varint(2)?;
                    Ok(())
                })?;

                // Login start
                write_packet(&mut stream, |buf| {
                    buf.write_u8_varint(0)?;
                    let uname = "McChatClient";
                    write_string(buf, uname)?;
                    Ok(())
                })?;

                read_packet(&mut stream, |packet| {
                    let id = packet.read_usize_varint()?;
                    assert!(id == 2);
                    let uuid = read_string(packet)?;
                    let username = read_string(packet)?;
                    Ok((uuid, username))
                })?
            }
            ServerType::Forge => {
                // Handshake

                write_packet(&mut stream, |buf| {
                    buf.write_u8_varint(0)?;
                    buf.write_u8_varint(5)?;
                    let addr = format!("{addr}\0FML\0");
                    buf.write_usize_varint(addr.len())?;
                    buf.write_all(addr.as_bytes())?;
                    buf.write_all(&port.to_le_bytes())?;
                    buf.write_u8_varint(2)?;
                    Ok(())
                })?;

                // Login start
                write_packet(&mut stream, |buf| {
                    buf.write_u8_varint(0)?;
                    let uname = "McChatClient";
                    write_string(buf, uname)?;
                    Ok(())
                })?;

                // Login Success
                let (uuid, username) = read_packet(&mut stream, |packet| {
                    let id = packet.read_usize_varint()?;
                    assert!(id == 2);
                    let uuid = read_string(packet)?;
                    let username = read_string(packet)?;
                    Ok((uuid, username))
                })?;

                // Server Register Channels
                let (_channels, channels_buf) =
                    read_plugin_msg(&mut stream, "REGISTER", |packet| {
                        Ok((
                            packet
                                .split(|b| *b == 0)
                                .map(|b| String::from_utf8(b.to_vec()).map_err(ChatError::from))
                                .collect::<Result<Vec<_>>>()?,
                            packet.to_vec(),
                        ))
                    })?;

                // Server Hello
                let _dim = read_plugin_msg(&mut stream, "FML|HS", |packet| {
                    let discriminant = VarintReader::read(packet)?;
                    assert_eq!(discriminant, 0);
                    let proto_ver = VarintReader::read(packet)?;
                    assert_eq!(proto_ver, 2);
                    let mut buf = [0u8; 4];
                    packet.read_exact(&mut buf)?;
                    let dim = i32::from_be_bytes(buf);
                    Ok(dim)
                })?;

                // Client Register Channels
                write_plugin_msg(&mut stream, "REGISTER", |body| {
                    body.write_all(&channels_buf)?;
                    Ok(())
                })?;

                // Client Hello
                write_plugin_msg(&mut stream, "FML|HS", |body| {
                    body.write_all(&[0x1])?;
                    body.write_all(&[0x2])?;
                    Ok(())
                })?;

                // Client Modlist
                write_plugin_msg(&mut stream, "FML|HS", |body| {
                    body.write_all(&[0x2])?;
                    let mods = mods.as_ref().unwrap();
                    body.write_usize_varint(mods.len())?;

                    for (id, ver) in mods {
                        write_string(body, id)?;
                        write_string(body, ver)?;
                    }
                    Ok(())
                })?;

                // Server Modlist
                let _mods = read_plugin_msg(&mut stream, "FML|HS", |packet| {
                    let discriminant = VarintReader::read(packet)?;
                    assert_eq!(discriminant, 2);
                    let num_mods = VarintReader::read(packet)?;

                    let mut mods = HashMap::new();
                    for _ in 0..num_mods {
                        let id = read_string(packet)?;
                        let ver = read_string(packet)?;
                        mods.insert(id, ver);
                    }
                    Ok(mods)
                })?;

                // HandshakeAck waitingseverdata
                write_plugin_msg(&mut stream, "FML|HS", |body| {
                    body.write_all(&[0xFF, 0x02])?;
                    Ok(())
                })?;

                // Server RegistryData
                read_plugin_msg(&mut stream, "FML|HS", |packet| {
                    let discriminant = VarintReader::read(packet)?;
                    assert_eq!(discriminant, 3);
                    let num_ids = packet.read_usize_varint()?;

                    for _ in 0..num_ids {
                        let _name = read_string(packet)?;
                        let _id = packet.read_usize_varint()?;
                    }

                    if packet.len() > 0 {
                        let len = packet.read_usize_varint()?;
                        for _ in 0..len {
                            read_string(packet)?;
                        }

                        let len = packet.read_usize_varint()?;
                        for _ in 0..len {
                            read_string(packet)?;
                        }
                    }
                    Ok(())
                })?;

                // Server HandshakeAck waitingack
                read_plugin_msg(&mut stream, "FML|HS", |packet| {
                    let discriminant = VarintReader::read(packet)?;
                    assert_eq!(discriminant, 255);
                    let phrase = VarintReader::read(packet)?;
                    assert_eq!(phrase, 2);
                    Ok(())
                })?;

                // HandshakeAck waitingsevercomplete
                write_plugin_msg(&mut stream, "FML|HS", |body| {
                    body.write_all(&[0xFF, 0x03])?;
                    Ok(())
                })?;

                // Server Mod registried
                while try_read_plugin_msg(&mut stream, |channel, packet| {
                    dbg!(channel);
                    if channel != "FML|HS" {
                        Ok(true)
                    } else {
                        let discriminant = VarintReader::read(packet)?;
                        assert_eq!(discriminant, 255);
                        let phrase = VarintReader::read(packet)?;
                        assert_eq!(phrase, 3);
                        Ok(false)
                    }
                })? {}

                // HandshakeAck complete
                write_plugin_msg(&mut stream, "FML|HS", |body| {
                    body.write_all(&[0xFF, 0x05])?;
                    Ok(())
                })?;

                (uuid, username)
            }
        };

        Ok(Server {
            stream,
            addr,
            port,
            lang,
            state: Connected {
                typ,
                description,
                uuid,
                username,
            },
        })
    }
}
