use std::{
    io::{Read as _, Write as _, stdin, stdout},
    net::{Shutdown, TcpStream},
};

use serde_json::json;
use varint_rs::{VarintReader, VarintWriter};

use crate::{
    Connected, Server,
    error::Result,
    readers::{read_packet, read_string},
    text::Message,
    writers::{write_packet, write_string},
};

pub enum Packet {
    KeepAlive([u8; 4]),
    Chat(Vec<Message>),
    Disconnect,
    Other(usize, Vec<u8>),
}

impl Server<Connected> {
    pub fn run(self) -> Result<()> {
        let Self {
            mut stream,
            lang,
            state: Connected { username, .. },
            ..
        } = self;
        let mut recv_stream = stream.try_clone().unwrap();
        std::thread::spawn(move || -> Result<()> {
            loop {
                let packet = read_packet(&mut recv_stream, parse_packet)?;
                match packet {
                    Packet::KeepAlive(id) => handle_keepalive(&mut recv_stream, id)?,
                    Packet::Chat(m) => {
                        for m in m {
                            println!("{}", m.format(&lang))
                        }
                    }
                    Packet::Disconnect => break,
                    Packet::Other(_, _) => {}
                }
            }
            Ok(())
        });
        for line in stdin().lines() {
            print!("\x1b[1F\x1b[0K");
            let line = &line?;
            write_packet(&mut stream, move |buf| {
                buf.write_usize_varint(0x01)?;
                write_string(buf, line)?;
                Ok(())
            })?;
        }
        stream.shutdown(Shutdown::Both)?;
        Ok(())
    }
}

fn parse_packet(packet: &mut &[u8]) -> Result<Packet> {
    let id = packet.read_usize_varint()?;
    match id {
        0x00 => {
            let mut id = [0u8; 4];
            packet.read_exact(&mut id)?;
            Ok(Packet::KeepAlive(id))
        }
        0x02 => {
            let msg = read_string(packet)?;
            let msg = Message::from_string(msg).unwrap();
            Ok(Packet::Chat(msg))
        }
        0x40 => Ok(Packet::Disconnect),
        i => Ok(Packet::Other(i, packet.to_vec())),
    }
}

fn handle_keepalive(stream: &mut TcpStream, id: [u8; 4]) -> Result<()> {
    let mut buf: Vec<u8> = Vec::new();
    buf.write_u8_varint(0)?;
    buf.write_all(&id)?;

    stream.write_usize_varint(buf.len())?;
    stream.write_all(&buf)?;
    Ok(())
}
