use std::{collections::HashMap, io::Write, net::TcpStream};

use serde_json::Value;
use varint_rs::{VarintReader, VarintWriter};

use crate::{
    Disconnected, Server, Status,
    error::ChatError,
    error::JsonError,
    error::Result,
    readers::{read_packet, read_string},
    writers::write_packet,
};

#[derive(Debug)]
pub enum ServerType {
    Vanilla,
    Forge,
}

impl Server<Disconnected> {
    pub fn status(self) -> Result<Server<Status>> {
        let Self {
            mut stream,
            addr,
            port,
            lang,
            ..
        } = self;

        // Handshake

        write_packet(&mut stream, |buf| {
            buf.write_u8_varint(0)?;
            buf.write_u8_varint(5)?;
            buf.write_usize_varint(addr.len())?;
            buf.write_all(addr.as_bytes())?;
            buf.write_all(&port.to_le_bytes())?;
            buf.write_u8_varint(1)?;
            Ok(())
        })?;

        write_packet(&mut stream, |buf| {
            buf.write_all(&[0x0])?;
            Ok(())
        })?;

        let json: Value = read_packet(&mut stream, |packet| {
            let id = packet.read_usize_varint()?;
            assert!(id == 0);
            let json = read_string(packet)?;
            serde_json::from_str(&json)
                .map_err(<JsonError as From<serde_json::Error>>::from)
                .map_err(ChatError::from)
        })?;

        let json = json
            .as_object()
            .ok_or(JsonError::ExpectedObject(json.clone()))?;
        let description = json
            .get("description")
            .ok_or(JsonError::ExpectedMember("description".to_string()))?
            .as_str()
            .ok_or(JsonError::ExpectedString)?
            .to_string();

        let (typ, mods) = match json.get("modinfo") {
            Some(info) => {
                if info
                    .as_object()
                    .ok_or(JsonError::ExpectedObject(info.clone()))?
                    .get("type")
                    .ok_or(JsonError::ExpectedMember("type".to_string()))?
                    .as_str()
                    .ok_or(JsonError::ExpectedString)?
                    == "FML"
                {
                    let mod_list = info
                        .get("modList")
                        .ok_or(JsonError::ExpectedMember("modList".to_string()))?;
                    let mod_list = mod_list.as_array().ok_or(JsonError::ExpectedArray)?;
                    let mut mods = HashMap::new();
                    for mod_ in mod_list {
                        let mod_ = mod_
                            .as_object()
                            .ok_or(JsonError::ExpectedObject(mod_.clone()))?;
                        let id = mod_
                            .get("modid")
                            .ok_or(JsonError::ExpectedMember("modid".to_string()))?;
                        let ver = mod_
                            .get("version")
                            .ok_or(JsonError::ExpectedMember("version".to_string()))?;
                        mods.insert(
                            id.as_str().ok_or(JsonError::ExpectedString)?.to_string(),
                            ver.as_str().ok_or(JsonError::ExpectedString)?.to_string(),
                        );
                    }
                    (ServerType::Forge, Some(mods))
                } else {
                    panic!()
                }
            }
            None => (ServerType::Vanilla, None),
        };

        stream.shutdown(std::net::Shutdown::Both)?;

        Ok(Server {
            stream: TcpStream::connect(format!("{addr}:{port}"))?,
            addr,
            port,
            lang,
            state: Status {
                typ,
                description,
                mods,
            },
        })
    }
}
