use serde_json::{Value, to_string};

use crate::lang::Lang;

#[derive(Debug)]
pub struct Message {
    content: MessageContent,
}

#[derive(Debug)]
#[allow(unused)]
pub enum MessageContent {
    Text(String),
    Translate {
        translate: String,
        fallback: Option<String>,
        with: Option<Vec<Message>>,
    },
    Score {
        name: String,
        objective: String,
    },
    Selector {
        selector: String,
        seperator: Value,
    },
    Keybind(String),
    Nbt {
        source: String,
        nbt: String,
        interpret: bool,
        seperator: String,
    },
}

impl Message {
    pub fn format(self, lang: &Lang) -> String {
        let Self { content } = self;
        match content {
            MessageContent::Text(t) => t,
            MessageContent::Translate {
                translate,
                fallback: _,
                with,
            } => {
                if let Some(template) = lang.get(translate.as_str()) {
                    let mut template = template.to_string();
                    if let Some(with) = with {
                        for subst in with.into_iter().map(|m| m.format(lang)) {
                            if template.contains("%s") {
                                template = template.replacen("%s", &subst, 1);
                            } else {
                                break;
                            }
                        }
                        template
                    } else {
                        template
                    }
                } else {
                    translate
                }
            }
            m => format!("{:?}", m),
        }
    }

    pub fn from_string(string: String) -> Option<Vec<Self>> {
        let json: Value = serde_json::from_str(&string).ok()?;
        Self::from_json(json)
    }

    fn from_json(message: Value) -> Option<Vec<Self>> {
        let message = message.as_object()?;

        let msg = match message.get("type") {
            Some(typ) => match typ.as_str()? {
                "text" => Some(Message {
                    content: MessageContent::Text(message.get("text")?.as_str()?.to_string()),
                }),
                "translatable" => Some(Message {
                    content: MessageContent::Translate {
                        translate: message.get("translate")?.as_str()?.to_owned(),
                        fallback: message
                            .get("fallback")
                            .map(|v| v.as_str().map(str::to_string))
                            .flatten(),
                        with: message
                            .get("with")
                            .map(|a| {
                                a.as_array().cloned().map(|a| {
                                    a.into_iter()
                                        .map(|v| {
                                            if let Some(s) = v.as_str() {
                                                vec![Message {
                                                    content: MessageContent::Text(s.to_string()),
                                                }]
                                            } else {
                                                Message::from_json(v).unwrap()
                                            }
                                            .into_iter()
                                        })
                                        .flatten()
                                        .collect()
                                })
                            })
                            .flatten(),
                    },
                }),
                "score" => {
                    let score = message.get("score")?.as_object()?;
                    Some(Message {
                        content: MessageContent::Score {
                            name: score.get("name")?.as_str()?.to_string(),
                            objective: score.get("objective")?.as_str()?.to_string(),
                        },
                    })
                }
                "selector" => Some(Message {
                    content: MessageContent::Selector {
                        selector: message.get("selector")?.as_str()?.to_string(),
                        seperator: message.get("seperator")?.clone(),
                    },
                }),
                "keybind" => Some(Message {
                    content: MessageContent::Keybind(message.get("keybind")?.as_str()?.to_string()),
                }),
                "nbt" => Some(Message {
                    content: MessageContent::Nbt {
                        source: message
                            .get(message.get("source")?.as_str()?)?
                            .as_str()?
                            .to_string(),
                        nbt: message.get("nbt")?.as_str()?.to_string(),
                        interpret: message
                            .get("interpret")
                            .map(|v| v.as_bool())
                            .flatten()
                            .unwrap_or(false),
                        seperator: message
                            .get("seperator")
                            .map(|v| v.as_str())
                            .flatten()
                            .unwrap_or(", ")
                            .to_string(),
                    },
                }),
                _ => return None,
            },
            None => {
                if let Some(txt) = message.get("text") {
                    Some(Message {
                        content: MessageContent::Text(txt.as_str()?.to_string()),
                    })
                } else if let Some(translate) = message.get("translate") {
                    Some(Message {
                        content: MessageContent::Translate {
                            translate: translate.as_str()?.to_owned(),
                            fallback: message
                                .get("fallback")
                                .map(|v| v.as_str().map(str::to_string))
                                .flatten(),
                            with: message
                                .get("with")
                                .map(|a| {
                                    a.as_array().cloned().map(|a| {
                                        a.into_iter()
                                            .map(|v| {
                                                if let Some(s) = v.as_str() {
                                                    vec![Message {
                                                        content: MessageContent::Text(
                                                            s.to_string(),
                                                        ),
                                                    }]
                                                } else {
                                                    Message::from_json(v).unwrap()
                                                }
                                                .into_iter()
                                            })
                                            .flatten()
                                            .collect()
                                    })
                                })
                                .flatten(),
                        },
                    })
                } else if let Some(score) = message.get("score") {
                    let score = score.as_object()?;
                    Some(Message {
                        content: MessageContent::Score {
                            name: score.get("name")?.as_str()?.to_string(),
                            objective: score.get("objective")?.as_str()?.to_string(),
                        },
                    })
                } else if let Some(selector) = message.get("selector") {
                    Some(Message {
                        content: MessageContent::Selector {
                            selector: selector.as_str()?.to_string(),
                            seperator: message.get("seperator")?.clone(),
                        },
                    })
                } else if let Some(keybind) = message.get("keybind") {
                    Some(Message {
                        content: MessageContent::Keybind(keybind.as_str()?.to_string()),
                    })
                } else if let Some(nbt) = message.get("nbt") {
                    Some(Message {
                        content: MessageContent::Nbt {
                            source: message
                                .get(message.get("source")?.as_str()?)?
                                .as_str()?
                                .to_string(),
                            nbt: nbt.as_str()?.to_string(),
                            interpret: message
                                .get("interpret")
                                .map(|v| v.as_bool())
                                .flatten()
                                .unwrap_or(false),
                            seperator: message
                                .get("seperator")
                                .map(|v| v.as_str())
                                .flatten()
                                .unwrap_or(", ")
                                .to_string(),
                        },
                    })
                } else {
                    None
                }
            }
        }?;
        let mut extra = if let Some(extra) = message.get("extra") {
            let extra = extra.as_array()?;
            let extra: Vec<Message> = extra
                .into_iter()
                .map(|v| {
                    if let Some(s) = v.as_str() {
                        vec![Message {
                            content: MessageContent::Text(s.to_string()),
                        }]
                    } else {
                        Message::from_json(v.clone()).unwrap()
                    }
                    .into_iter()
                })
                .flatten()
                .collect();
            extra
        } else {
            Vec::new()
        };
        extra.push(msg);
        Some(extra)
    }
}
